/**************

 XPATCH
  Example of how this program is to be used:
  
     rm -rf /tmp/.wine-*
     LD_LIBRARY_PATH=/usr/lib/wine/lib \
     LD_PRELOAD=~bisqwit/nes/xgetimage/xpatch.so \
     /usr/lib/wine/bin/wine-pthread fam/FAMBLIP.EXE tinytoon.nes
  
  It logs video. Requires MEncoder.

 If USE_MEMFILE and USE_SHM are defined

   TODO: Document
 
 If USE_MEMFILE is defined but USE_SHM is not
   TODO: Document
 
 If USE_MEMFILE is not defined
   TODO: Document


****/

#include <dlfcn.h>
#include <X11/Xlib.h>
#include <fcntl.h>
#include <unistd.h>

#include <cstdio>
#include <algorithm>

#include "avifun.hh"

#define CALL_OLD

//#define SEND_LOGO

#ifdef USE_MEMFILE
 #ifdef USE_SHM
  #include <sys/ipc.h>
  #include <sys/shm.h>
 #else
  #include <sys/mman.h>
 #endif
#endif
#ifdef SEND_LOGO
 #include "tgaimage.hh"
 #include <math.h>
 #include <vector>
void CombineImage(TGAimage& target, const TGAimage& second, 
                  int xpos, int ypos)
{
/*
    printf("Bitness = %d / %d\n",
        target.GetPixBitness(),
        second.GetPixBitness());
*/
    for(unsigned y=0; y<second.GetYdim(); ++y)  
    {
        int desty = ypos + y;

        if(desty < 0 || desty >= target.GetYdim()) continue;
        
        for(unsigned x=0; x<second.GetXdim(); ++x)
        {
            int destx = xpos + x;
 
            if(destx < 0 || destx >= target.GetXdim()) continue;
            
            unsigned point = second.Point(x, y);
            
            if((point >> 24) < 128) continue;
/*
            printf("%08X\n", point);
*/
            target.PSet(destx, desty, point);
        }
    }
}
#endif

using std::fprintf;

#define XPutImage_f_params \
    Display*         display , \
    Drawable         d , \
    GC           gc ,    \
    XImage*      image , \
    int          src_x , \
    int          src_y , \
    int          dest_x , \
    int          dest_y , \
    unsigned int     width , \
    unsigned int     height 

int XPutImage(XPutImage_f_params)
{
    /* This object will load the _real_ XPutImage
     * symbol and enable calling it
     */
    static struct caller
    {
        int (*orig) (XPutImage_f_params);
    public:
        caller()
        {
            void *f = dlopen("libXext.so", RTLD_NOW);
            
            orig = (int(*)(XPutImage_f_params))dlsym(f, "XPutImage");
            
            dlclose(f);
        }
    } caller;
    
    static struct sender
    {
#ifndef USE_SHM
        int fd;
#endif
#ifdef USE_MEMFILE
        char *buffer;
        unsigned size;
#else
        bool first;
#endif
    public:
#ifdef USE_MEMFILE
        sender() : buffer(NULL), size(256*240*2)
#else
        sender() : first(true)
#endif
        {
            /* Prevent the subprocesses getting the same
             * measures as this victim program did
             */
            unsetenv("LD_PRELOAD");
            
#ifdef USE_MEMFILE
 #ifdef USE_SHM
            int fd;
 #endif
            fd = open(MEMORY_FILE.c_str(),
                      O_RDWR | O_CREAT | O_TRUNC, 0666);
            if(fd == -1)
            {
                perror(MEMORY_FILE.c_str());
            }
            
 #ifdef USE_SHM
            int shmid = shmget(IPC_PRIVATE, size, IPC_CREAT | 0666);
            if(shmid == -1)
            {
                perror("shmget");
            }
            char Buf[64];
            sprintf(Buf, "%d", shmid);
            write(fd, Buf, strlen(Buf));
            close(fd);
            
            buffer = (char*) (shmat(shmid, NULL, 0));
 #else
            ftruncate(fd, size);
            
            buffer = (char*) (mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0));
            if(!buffer || buffer == (char*)(MAP_FAILED))
            {
                perror("mmap");
            }
 #endif
#else
            OpenAVI();
#endif
        }
        ~sender()
        {
#ifdef USE_MEMFILE
 #ifdef USE_SHM
            shmdt(buffer);
 #else
            close(fd);
            munmap(buffer, size);
 #endif
            unlink(MEMORY_FILE.c_str());
#else
            CloseAVI();
#endif
        }
        void Send(const char *imagedata, unsigned width, unsigned height)
        {
#ifdef USE_MEMFILE
            std::copy(imagedata, imagedat + width*height*2,
                      buffer);
#else
            if(width != 256 || height < 224 || height > 240) return;
#ifdef SEND_LOGO
            if(first)
            {
                SendLogo(width,height);
            }
#endif
            SendFrame((const unsigned char*) imagedata, width, height, first);
            first = false;
#endif
        }
#ifdef SEND_LOGO
        void SendLogo(unsigned width,unsigned height)
        {
            #define B "/shares/home/bisqwit/nes/xgetimage/logo/"
            const char *background =
                width==320 ? B"logop0b.tga"
              : height==240 ? B"logop240.tga"
              : B"logop0.tga";
            TGAimage viiva(B"logop1.tga");
            TGAimage flat1(B"logop2.tga");
            TGAimage flat2(B"logop3.tga");
            TGAimage ukko(B"logop4.tga");
            #undef B
            
            const int destpos_x = 200*width/256;
            const int destpos_y = 190*height/224;
            
            const int length_total = 60;
            const int length_drop  = 35;
            const double accel = 1.2;
            int start_offs = -40;
            
            for(int frame = 0; frame < length_total; ++frame)
            {
                TGAimage image(background);
                
                if(frame < length_drop)
                {
                    int y = pow(frame, accel) * (destpos_y-start_offs)
                                              / pow(length_drop+1, accel)
                                              + start_offs;
                    CombineImage(image, viiva, destpos_x + 4, y);
                }
                else if(frame == length_drop
                     || frame == length_drop+1
                     || frame == length_drop+2
                     || frame == length_drop+3
                     || frame == length_drop+4
                     || frame == length_drop+9
                     || frame == length_drop+10
                     || frame == length_drop+11
                     || frame == length_drop+12)
                {
                    CombineImage(image, flat1, destpos_x, destpos_y);
                }
                else if(frame == length_drop+4
                     || frame == length_drop+5
                     || frame == length_drop+6
                     || frame == length_drop+7
                     || frame == length_drop+8)
                {
                    CombineImage(image, flat2, destpos_x, destpos_y+13);
                }
                else 
                {
                    CombineImage(image, ukko, destpos_x, destpos_y);
                }
                
                image.setboxsize(width,height);
                image.setboxperline(1);
                
                const std::vector<unsigned char>& data = image.GetData();
                std::vector<unsigned short> result(width*height);
                for(unsigned pos=0, max=result.size(); pos<max; ++pos)
                {
                    unsigned B = data[pos*3+0];
                    unsigned G = data[pos*3+1];
                    unsigned R = data[pos*3+2];
                    result[pos] = ((B*31/255)<<0)
                                | ((G*63/255)<<5)
                                | ((R*31/255)<<11);
                }
                SendFrame((const unsigned char*)&result[0], width,height, first);
                first=false;
            }
        }
#endif
    } sender;

    //fprintf(stderr, "Put(%u,%u) -> %p - %u\n", width,height, gc, gc->gid);
    
    sender.Send(image->data, width, height);

/*
    fprintf(stderr, "width=%u height=%u src=(%u,%u) dest=(%u,%u)\n",
       width,height,
       src_x,src_y,
       dest_x,dest_y);
*/
    

#ifdef CALL_OLD
    return caller.orig(display,d,gc,image,src_x,src_y,dest_x,dest_y,width,height);
#endif
    
    return 0;
}
